function [Yhat, Usample]=active_learning(ml_algo, X, known_idx, Ytr, nte)
%[Yhat, Usample]=active_learning(ml_algo, X, known_idx, Ytr, nte)
% Make predictions and rank samples according to uncertainty.
% Inputs:
% ml_algo       -- Name of a learning algorithm (e.g. a CLOP object)
% X             -- Data matrix
% known_idx     -- Indices of the known labels in the raws of X
% Ytr           -- Known labels (in the order of known_idx)
% nte           -- Number of test examples (to be taken from the end of the data matrix)
% Returns:
% Yhat          -- Label prediction scores (larger values for positive class examples)
% Usample       -- A sorted list of the raw numbers of X, most desirable labels come first

% Author: Isabelle Guyon -- November 2009 -- causality@clopinet.com

    if strcmp(ml_algo, 'zarbir')
        random_order=1;
    else
        random_order=0;
    end
    if strcmp(ml_algo, 'zarbip') || strcmp(ml_algo, 'zarbir')
       ml_algo='zarbi';
    end
    global mymodel                              % Custom CLOP model
    [P, N]=size(X);
    Yhat=zeros(P, 1);                           % Predictions on all samples
    Usample=[];                                 % Uncertain sample list
    train_num=P-nte;                            % Number of training examples
    test_idx=(train_num+1:P)';                  % Indices of test examples
    unknown_idx=setdiff((1:train_num)', known_idx); % Indices of unknown examples
    Xtr=X(known_idx, :);                        % Training data
    te_idx=[unknown_idx; test_idx];             % Indices of test examples
    Xte=X(te_idx, :);                           % Test data
    negidx=find(Ytr==-1, 1);                    % Negative examples
	if strcmp(ml_algo, 'rand') || strcmp(ml_algo, 'opt') 
        Yhat=rescale_resu(randn(size(X, 1), 1)); % Random predictions
    elseif ~exist(ml_algo)
        Yhat=[];
        fprintf('==> ERROR: the model %s could not be found\n', ml_algo);
        return
    elseif isempty(negidx) % No example of the negative class
        % Use as prediction a ranking according to Euclidean distance
        D2=sum((X-Xtr(ones(size(X,1),1), :)).^2, 2);
        % Exponentiate so the known label has value 1 and the other values
        % are smaller
        Yhat=exp(-D2);
    else
        % Make predictions according to ml_algo
        Dtr=data_struct(Xtr, Ytr);
        Dte=data_struct(Xte);
        if exist('bias', 'file') == 2
            [tr_result, trained_ml_algo]=train(chain({eval(ml_algo), bias}), Dtr);
        else
            [tr_result, trained_ml_algo]=train(eval(ml_algo), Dtr);
        end
        % Compute the training AUC just to check
        train_auc=auc(tr_result.X, tr_result.Y);
        if ~isempty(train_auc), fprintf('==> Training AUC = %g\n', train_auc); end
        if ~isempty(te_idx)
            te_result=test(trained_ml_algo, Dte);
            % Rescale the results so the most confident predictions are +-1
            % to make sure that the known labels have extreme values
            Yhat(te_idx)=rescale_resu(te_result.X);
        end
    end
    % Predict correctly the known labels
    Yhat(known_idx)=Ytr;
    %%%%%       Uncertainty sampling algorithm     %%%%%
    % Sort the samples so the most uncertain come first
    if ~random_order
        [xx, sidx]=sort(abs(Yhat(unknown_idx)));
    else
        sidx=randperm(length(unknown_idx));
    end
    Usample=unknown_idx(sidx);   
    %%%%%       Change to something more clever    %%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
return