%==========================================================================
% DBN Discrete Bayesian network model         
%========================================================================== 
% a=DBN(hyper) 
% DBN is derived from "model".
% To derive Bayesian networks from that class
% 1) Create a new directory: @mydbn and create in it a file mydbn.m 
% containing the 2 lines:
%       classdef mydbn < DBN
%       end
% 2) Add in the same directory a file mydbn_bnet.m that constructs a bnet 
% or a file mydbn_bnet.mat with a saved bnet object.
% See the example below a=sprinkler_bnet(a) for an example and
% the documentation of K. Murphy's Bayesian network toolbox. 

% Uses K. Murphy's Bayesian network toolbox.
% Isabelle Guyon -- isabelle@clopinet.com -- October 2009

classdef DBN < model
    
    properties (SetAccess = protected)
        current_budget=0; % Internally keeps track of budget
        bnet=[];          % Murphy style Bayesian network
    end
    
     % <<-------------public methods----------------->> 
    methods
        %%%%%%%%%%%%%%%%%%%
        %%% CONSTRUCTOR %%%
        %%%%%%%%%%%%%%%%%%%
        function a=DBN(hyper)
            if nargin<1, hyper={}; end
            % Call the constructor of the parent class (call init)
            a=a@model(hyper); 
            % Initialize budget
            a.current_budget=a.initial_budget;
            % Run only for debug purpose (slow)
            debug=0;
            if debug
                OK=check_budget(a);
                OK=OK && check_net(a);
                if ~OK, error('Budget or structure error'); end
            end
        end    
        
        %%%%%%%%%%%%%%%%%%%  
        function dag=dag(a)
            dag=a.bnet.dag;
        end
        function v=var_name(a, num)
            if any(a.target==0), add1=1; else add1=0; end
            if iscell(a.bnet.names)
                v=a.bnet.names;
            else
                v=struct(a.bnet.names).keys';
            end
            if nargin>1,
                 v=v{num+add1};
            end
        end
        function v=var_type(a, num)
            if any(a.target==0), add1=1; else add1=0; end
            v=cell(1,varnum(a));
            vr=var_max(a)-var_min(a);
            bi=find(vr==1 | var_min(a)<0);
            [v{:}]=deal('categorical');
            [v{bi}]=deal('binary');
            if nargin>1, 
                v=v{num+add1};
            end
        end
        function v=var_min(a, num)
            if any(a.target==0), add1=1; else add1=0; end
            v=ones(1,length(a.bnet.node_sizes));
            if nargin>1, 
                v=v(num+add1);
            end
        end  
        function v=var_max(a, num)
            if any(a.target==0), add1=1; else add1=0; end
            v=a.bnet.node_sizes;      
            if nargin>1, 
                v=v(num+add1);
            end
        end   
    end
    
    % <<-------------protected methods----------------->> 
    methods (Access = protected)
        function a=init(a)
            a.time_dept=0; % Time independent network
            % Set properties
            a.contact_name='Isabelle Guyon';
            a.contact_url='http://clopinet.com/isabelle/';
            a.title='Discrete Bayesian Network';
            a.authors='Causality Workbench Team';
            a.keywords='bayesian.network';
            % Load or create a network
            bnetname=[class(a) '_bnet'];
            if ~(exist([whereis(a) bnetname '.mat'], 'file')==2)
                if ~(exist([whereis(a) bnetname '.m'], 'file')==2)
                    error('DBN::init: No %s found\n', bnetname);
                else
                    fprintf('DBN::init: Initializing %s\n', bnetname);
                    a=feval(bnetname, a);
                    save_bnet(a, [whereis(a) bnetname '.mat']);
                end
            else
                fprintf('DBN::init: Reloading bnet %s\n', bnetname);
                a=load_bnet(a, [bnetname '.mat']);
            end
            % Set the variable properties (observable, actionable, et.)
            a=default_var_properties(a);
            % Add names, if none
            if isempty(a.bnet.names),
                fprintf('DBN::init: Adding names to %s\n', bnetname);
                for k=1:varnum(a)
                    a.bnet.names{k}=['var' num2str(k)];
                end
                if any(a.target==0), a.bnet.names=['target' a.bnet.names{k}]; end
            end
        end
        
        function a=default_var_properties(a)
            N=length(a.bnet.dag);
            % Set the network properties
            if isempty(a.observable)
                a.observable=1:N;       % All variables are observable
            end
            if isempty(a.actionable)
                a.actionable=1:N;       % All variables are actionable
            end
            if isempty(a.target)
                a.cost_per_target_observation=0; % Of course since there is no target
            end
        end
        
        function a=load_bnet(a, filename)
            % Reload a network from a file
            warning off
            load(filename);
            warning on
            a.bnet=bnet;
        end
        
        function a=save_bnet(a, filename)
            % Save a network to a file
            bnet=a.bnet;
            save(filename, 'bnet');
        end
        
        function a=sprinkler_bnet(a)
            % Create a small example network (After K. Murphy)
            N = 4; 
            dag = zeros(N,N);
            C = 1; S = 2; R = 3; W = 4;
            dag(C,[R S]) = 1;
            dag(R,W) = 1;
            dag(S,W)=1;

            % Declare discrete nodes
            discrete_nodes = 1:N;
            node_sizes = 2*ones(1,N); 

            % Create the net
            bnet = mk_bnet(dag, node_sizes, 'names', {'cloudy','sprinkler','rain','wet'}, 'discrete', 1:4);
            C = bnet.names('cloudy'); 
            S = bnet.names('sprinkler'); 
            R = bnet.names('rain'); 
            W = bnet.names('wet'); 

            bnet.CPD{C} = tabular_CPD(bnet, C, [0.5 0.5]);
            bnet.CPD{R} = tabular_CPD(bnet, R, [0.8 0.2 0.2 0.8]);
            bnet.CPD{S} = tabular_CPD(bnet, S, [0.5 0.9 0.5 0.1]);
            bnet.CPD{W} = tabular_CPD(bnet, W, [1 0.1 0.1 0.01 0 0.9 0.9 0.99]);
            a.bnet=bnet;
            
            % Variable properties
            a.actionable = [2 4];
        end % sprinkler_bnet
        
        function OK=check_budget(a)
        %OK=check_budget(a)
        % Verify by processing actual queries that the budget was computed right
            OK=1;
            fprintf('DBN::check_budget: == Compute cost of labeled training set ==\n')
            train_cost=cost(process_query(a, query('TRAIN 2'))); % 2 means labeled
            fprintf('DBN::check_budget: == Compute cost of test set ==\n')
            test_cost=cost(process_query(a, query('TEST')));
            initial_budget=train_cost+test_cost;
            if a.initial_budget~=initial_budget
                OK=0;
                fprintf('DBN::check_budget: ERROR, budgets do not match, true initial budget=%d+%d=%d\n', train_cost,test_cost,initial_budget);
            else
                fprintf('DBN::check_budget: budget OK\n');
            end
        end
        
    end
end
