function Xresu=draw_sample(a, Xinit, Xmanip, p)
%Xresu=draw_sample(a, Xinit, Xmanip, p)
% Lower level sampling function for discrete Bayesian networks.
% Xinit is the matrix of initialization values
% Xmanip is the matrix of manipulations
% NaN or 0: no evidence (missing value)
% Positive values: some evidence (binary values are 1, 2)
% If Xmanip and Xinit are not given, draw p samples from bnet and returns
% a matrix of p lines with variables in columns.

% Isabelle Guyon -- isabelle@clopinet.com -- May/Oct 2009 

debug=0; % Flag to perform additional tests

% === Verify the arguments ===
if nargin<4, p=[]; end
if nargin<3, Xresu=Xinit; return; end
if isempty(Xmanip) && isempty(p), Xresu=Xinit; return; end

bnet=a.bnet;
n=length(bnet.dag);
if ~isempty(Xinit)
    [p1, n1]=size(Xinit); 
    if n~=n1, error('DBN::draw_sample: inconsistent dimensions'); end
end
if ~isempty(Xmanip)
    [p2, n2]=size(Xmanip); 
    if n~=n2, error('DBN::draw_sample: inconsistent dimensions'); end
end

if ~isempty(Xinit) && ~isempty(Xmanip)
    if p1~=p2, error('DBN::draw_sample: inconsistent dimensions'); end
    p=p1;
end

% === End verify the arguments ===

% Display the list of initialized and manipulated variables
if iscell(bnet.names)
    names=bnet.names;
else
    names=struct(bnet.names).keys;
end
if ~isempty(Xinit) && ~isempty(Xmanip)
    if debug
        % The way we set up things, we always initialize everything when we
        % manipulate
        init_list=find(any(Xinit));
        fprintf('DBN::draw_sample: Initialization list: ');
        if isempty(init_list)
            fprintf('empty');
        else
            for k=1:length(init_list), fprintf('%s, ', names{init_list(k)}); end
        end
        fprintf('\n');
    end
    manip_list=find(any(Xmanip));
    fprintf('DBN::draw_sample: Manipulated list: ');
    if isempty(manip_list)
        fprintf('empty');
    else
        for k=1:length(manip_list), fprintf('%s, ', names{manip_list(k)}); end
    end
    fprintf('\n');
else
    fprintf('DBN::draw_sample: Drawing premanipulation values\n');
end

% Loop over samples to be drawn
Xresu=zeros(p,n);
for k=1:p
    xinit=NaN*zeros(1,n); xmanip=NaN*zeros(1,n); 
    if ~isempty(Xinit), xinit=Xinit(k,:); end
    if ~isempty(Xmanip), xmanip=Xmanip(k,:); end
    Xresu(k, :) = sample_bnet(a, xinit, xmanip); % <== Lower level sampling function (inspired by K. Murphy's code by re-written to accomodate both initialization and manipulation)
end

if debug
    % Verify that the CPDs for the drawn samples are correct
    if isempty(Xinit) && isempty(Xmanip), verify_cpd(a, Xresu); end
end

% Verify that the initialized root nodes are unchanged
root_idx=root_list(a);
is_ok=1;
if ~isempty(Xinit)
    is_ok=all(all(Xresu(:, root_idx)==Xinit(:, root_idx)));
end
if ~is_ok
    fprintf('DBN::draw_sample: Warning: Initial root values changed after DBN evaluation!\n');
end

% Verify that the clamped nodes remain unchanged
clamped_idx=find(Xmanip~=0 & ~isnan(Xmanip));
if ~isempty(clamped_idx)
    is_ok=all(Xresu(clamped_idx)==Xmanip(clamped_idx));
    if ~is_ok
        fprintf('DBN::draw_sample: Warning: Evidence changed after manipulation!\n');
    end
end



