function [q, a] = sample(a, q)
%q = sample(a, q)
%[q, a] = sample(a, q)
% Generates data from the model.
% Inputs:
% a -- A model object.
% q -- A query object or a sample number (if observational data are
% requested only).
% Returns:
% q -- The query object, now containing the answer.
% a -- Optionally get back "a" with updated budget.

% Isabelle Guyon -- May/Oct 2009 -- isabelle@clopinet.com

% Important: update the number of variables queried
q=quantity(q, a);
% and verify the range of the variables
q=verify_range(q, a);

% PREMANIPULATION DATA
% ====================
has_target=any(a.target==0);
if is_survey(q) && ~has_target
    fprintf('WARNING: DBN::sample: Survey data not supported\n');
    return
end
tag='';
if is_test(q)
    tag='test'; 
elseif is_train(q) || is_survey(q)
    tag='train';
end

% Draw or reload initial values
Xinit=reload_sample(a, tag); % Will be [] if tag==''
if isempty(Xinit)
    % Draw natural unmanipulated samples [premanip values]
    Xinit=draw_sample(a, [], [], samplenum(q));  
    save_sample(a, tag, Xinit); % Save only if the tag is not empty
end

% Record as premanipval the values requested
nidx0=sanitized_idx(a, premanipvar(q), 'observable');
pidx=sample(q);
if isempty(pidx), pidx=1:size(Xinit, 1); end
if ~isempty(nidx0)
    q=premanipvar(q, nidx0-has_target); 
    q=premanipval(q, Xinit(pidx, nidx0)); 
end

% MANIPULATION DATA
% =================
if is_test(q)
    % Return the manipulations to be performed
    % The test set 0 is observational data
    % But we SET manipvar and manipval (this is our query to the participants)
    q=manipulate(a, q);
end
    
% POSTMANIPULATION DATA
% =====================
newt=is_test(q);
% First time we get test data: compute the post-manip results
fn=[class(a) '_test' num2str(newt) '.mat'];
if exist(fn, 'file')==2 || isempty(manipvar(q))
    newt=0;
end

% Create a manipulation matrix
[nidx1, idx1]=sanitized_idx(a, manipvar(q), 'actionable');
nidx2=sanitized_idx(a, postmanipvar(q), 'observable');

do_manip = ~isempty(nidx1)&&~isempty(nidx2)&&~is_test(q);
N = varnum(a); %  Total number of variables (excluding the target)
Xresu=[];
if do_manip || newt
    Xmanip=NaN*zeros(size(Xinit));
    mval=manipval(q); 
    mval=mval(:,idx1);
    Xmanip(:, nidx1)=mval; 
    Xresu=draw_sample(a, Xinit, Xmanip);
    q=manipvar(q, idx1-has_target);
    q=manipval(q, mval);
end

if newt
    save_sample(a, 'test1', Xresu); % A single test set for DBN
elseif do_manip % No postmanip observation if no manipulation
    if size(nidx2,1)==1
        Xresu=Xresu(:, nidx2);
    else % postmanipvar change for each sample
        Y=zeros(size(nidx2));
        for k=1:size(nidx2,1)
            ii=nidx2(k,:);
            Y(k)=Xresu(k,ii);
        end
        Xresu=Y;
    end
    q=postmanipvar(q, nidx2-has_target);
    q=postmanipval(q, Xresu);
end    

% COMPUTE THE COST OF THE QUERY
% =============================
q=varnum(q, N); % Return the variable number fr information
[q, a]=compute_cost(a, q);

return


function [nidx, idx]=sanitized_idx(a, nidx0, type)

if strcmp(type, 'observable')
    gidx=a.observable;
else
    gidx=a.actionable;
end

if size(nidx0,1)>1
    idx=1:varnum(a);
    nidx=nidx0;
    return
end

nidx=nidx0;
has_target=any(a.target==0);
if has_target, add1=1; else add1=0; end
cadd1=0;
if add1==0 && any(nidx==0), 
    fprintf('WARNING: DBN::sample: zero index found in a query to a network having no target,adding 1\n');
    cadd1=1;
end
nidx=nidx+add1+cadd1; %nidx: 0-based indices when target is present
gidx=gidx+add1;
no_good=setdiff(nidx, gidx);
if ~isempty(no_good)
    fprintf('WARNING: DBN::sample: Forbidden index ignored\n');
end
good_idx=setdiff(nidx, no_good);
idx=zeros(size(good_idx));
for k=1:length(idx)
    idx(k)=find(nidx==good_idx(k));
end
nidx=good_idx;