%==========================================================================
% Active Learning model         
%==========================================================================
%
% a=al_model(hyper) 
% al_model is derived from "model".
% To derive an active learning model from this class
% 1) Create a new directory: @my_al_model and create in it a file my_al_model.m 
% containing the 2 lines:
%       classdef my_al_model < al_model
%       end
% 2) Add in the same directory a my_al_model.config file and a task_n_pricing.m. 

% Isabelle Guyon -- isabelle@clopinet.com -- November 2009

classdef al_model < model
    
    properties (SetAccess = protected)
        current_budget=0; % Internally keeps track of budget
        data=[];          % Data matrix
        label=[];         % Label truth values
        dataid={};        % Names of the columns of the labels matrix
        featid={};        % Names of the columns of the data matrix
        labelid={};       % Ids of the samples
        targetid=[];      % ID of the selected target (personal to the participant)
        adjustment_weights=[]; % Weighting factors for the tasks
        queried_samples=[]; % List of samples already queried
        cash_history=[];  % Amounts of cash spent
        auc_history=[];   % AUC scores
        ebar_history=[];  % Error bars
        current_score=[]; % Area under the learning curve
        axis_scale='linear'; % Scale of the learning curve axis ('linear', 'sqrt', or 'log')
        seed=[];
        auc_wo_queried_samples=1;   % Remove the known labels when computing the auc
    end
    
     % <<-------------public methods----------------->> 
    methods
        %%%%%%%%%%%%%%%%%%%
        %%% CONSTRUCTOR %%%
        %%%%%%%%%%%%%%%%%%%
        function a=al_model(hyper)
            if nargin<1, hyper={}; end
            % Call the constructor of the parent class (call init)
            a=a@model(hyper); 
            % Initialize budget (we pay for the seed)
            a.current_budget=a.initial_budget-1;
            a.cash_history=1; 
            a.queried_samples=a.seed;
        end    
        %%%%%%%%%%%%%%%%%%%
        
        function save(L, filename)
            %save(L, filename)
            
            % Clear the data (too big)
            L.data=[];         
            %L.label=[];         
            %L.dataid=[];       
            %L.featid=[];       
            %L.labelid=[];      
            % save the model
            save(filename, 'L');
        end
        
        function s=sample_num(a)
            if ~isempty(a.label)
                s=size(a.label, 1);
            else
                s=a.train_num+a.test_num;
            end
        end
        
        function v=var_name(a, num, is_label)
            if nargin<2, num=[]; end
            if nargin<3, is_label=0; end
            if is_label,
                v=a.labelid;
                add1=0;
            else
                v=['target' a.featid];
                add1=1;
            end
            if ~isempty(num),
                 v=v{num+add1};
            end        
        end
        function v=var_type(a, num, is_label)
            if nargin<2, num=[]; end
            if nargin<3, is_label=0; end
            if is_label,
                M=a.label;
                add1=0;
            else
                M=a.data;
                add1=1;
            end
            v={};
            for k=1:size(M, 2)
                u=unique(M(:,k));
                if length(u)<3
                    v{k}='binary';
                elseif length(u)<20
                    v{k}='categorical';
                else
                    v{k}='continuous';
                end
            end
            if ~is_label,
                v = ['binary' v];
            end
            if ~isempty(num),
                v=v{num+add1};
            end
        end
        
        function v=var_min(a, num, is_label)
            if nargin<2, num=[]; end
            if nargin<3, is_label=0; end
            if is_label,
                M=a.label;
                add1=0;
            else
                M=a.data;
                add1=1;
            end
            v=min(M);
            v=[-1 v];
            if ~isempty(num),
                v=v(num+add1);
            end
        end  
        function v=var_max(a, num, is_label)
            if nargin<2, num=[]; end
            if nargin<3, is_label=0; end
            if is_label,
                M=a.label;
                add1=0;
            else
                M=a.data;
                add1=1;
            end
            v=max(M);  
            v=[1 v];
            if ~isempty(num),
                v=v(num+add1);
            end
        end   
        function v=sparsity(a)
            v=1-(nnz(a.data)/numel(a.data));
        end
        function v=frac_pos(a)
            N=size(a.label,1);
            for k=1:size(a.label, 2)
                v(k)=length(find(a.label(:,k)==1))/N;
            end
        end
        function v=frac_missing(a)
            v=length(find(isnan(a.data)))/numel(a.data);
        end
        
        function c=data_profile(a, show)
            if nargin<2, show=0; end
            header={'Dataset', 'Domain', 'Feat. type', 'Feat. num.', 'Sparsity (%)', 'Missing (%)', 'Label', 'Positive labels (%)', 'Train num.', 'Test num.', 'Axis scale', 'Seed'};
            c{1}=upper(class(a)); % Dataset
            c{2}=a.title; % Domain
            v=var_type(a);
            u=unique(v(2:length(v)));
            if length(u)==1
                c(3)=u; % Feat. type
            else
                c{3}='mixed';
            end
            c{4}=num2str(length(a.featid)); % Feat. num
            c{5}=num2str(round(sparsity(a)*10000)/100); % Sparsity
            c{6}=num2str(round(frac_missing(a)*1000000)/10000); % Frac. missing
            v=var_type(a, [], 1);
            u=unique(v);
            if length(u)==1
                c(7)=u; % Label type
            else
                c{7}='mixed';
            end
            c{8}=num2str(round(median(frac_pos(a))*10000)/100); % Frac. pos. lbl
            c{9}=num2str(a.train_num); % Training set size
            c{10}=num2str(a.test_num); % Test set size
            c{11}=a.axis_scale; % Axis scale
            c{12}=num2str(a.seed); % Seed
            c=[header ;c];
            if show,
                for i=1:size(c, 1)
                    for j=1:size(c, 2)
                        fprintf('%s\t', c{i, j});
                    end
                    fprintf('\n');
                end
                fprintf('\n');
                label_profile(a, [], 1);
                fprintf('\n');
                var_profile(a, [], [], 1);
            end   
        end % data_profile
        
        function c=var_profile(a, num, as_xml, show)
            % a=var_profile(a)
            % c=var_profile(a, num)
            % Function that returns the profile of variables 
            % name, accessibility (observable, actionable, unobservable),
            % type (catagorical, binary, ordinal, continuous), min value,
            % max value. Note: unobservable variables are neither
            % observable nor actionable. Actionable variables are
            % observable.
            if nargin<2, num=[]; end
            if nargin<3, as_xml=0; end
            if nargin<4, show=0; end
            if ~isempty(a.data)
                c=var_profile@model(a, num, as_xml, show);
            else
                c=a;
            end
        end
        
        function c=label_profile(a, num, show)
            % c=label_profile(a)
            % c=label_profile(a, num)
            % c=label_profile(a, num, shown)
            % Function that returns the profile of the labels
            if nargin<2, num=[]; end
            if nargin<3, show=0; end
            header={'Index', 'Name', 'Access', 'Type', 'Min', 'Max', 'FracPos (%)'};
            name=a.labelid;
            N=length(name);
            for k=1:N
                access{k}='observable';
            end
            type=var_type(a, [], 1);
            min=var_min(a, [], 1);
            max=var_max(a, [], 1);
            fracpos=frac_pos(a);
            c=cell(N,6);
            for k=1:N
                c{k,1}=num2str(k);
                c{k,2}=name{k};
                c{k,3}=access{k};
                c{k,4}=type{k};
                c{k,5}=num2str(min(k));
                c{k,6}=num2str(max(k));
                c{k,7}=num2str(round(10000*fracpos(k))/100);
            end
            c=[header; c];
            if ~isempty(num),
                n=num+1;
                c=c(n,:); 
            end 
            if show,
                for i=1:size(c, 1)
                    for j=1:size(c, 2)
                        fprintf('%s\t', c{i, j});
                    end
                    fprintf('\n');
                end
            end
        end
    end % label_profile
    
    % <<-------------protected methods----------------->> 
    methods (Access = protected)
        function a=init(a)
            a.time_dept=0; % Time independent network
            % Load or create a model
            a=load_sample(a);
            % Set the variable properties (observable, actionable, et.)
            a=default_var_properties(a);
            % Consistency of the sample number
            if ~a.auc_wo_queried_samples
                a.train_num=[];
            end
            if isempty(a.train_num)
                a.train_num=a.sample_num;
            end
            if a.train_num > a.sample_num
                error('Too many training samples');
            else
                a.test_num=a.sample_num-a.train_num;
            end
            % Target (target is one of the labels drawn at random but is
            % always numbered 0).
            a.target=0;
            if isempty(a.targetid) 
                if ~isempty(a.labelid)
                    a.targetid=randi(length(a.labelid), 1, 1);
                else
                    a.targetid=1;
                end
            else
                clear a.data; % No use
            end
            % Find the seed (first positive example)
            pidx=find(a.label(:,a.targetid)==1);
            a.seed=pidx(1);
        end
        
        function a=default_var_properties(a)
            N=size(a.data, 2);
            % Set the network properties
            if isempty(a.observable)
                a.observable=0:N;       % All variables are observable
            end
            if isempty(a.actionable)
                a.actionable=[];       % No variable actionable
            end
        end
        
        function a=load_config(a)
            % Load hyperparamters from configuration file
            configfile=[class(a) '_config.txt'];
            if (exist(configfile, 'file')==2)
                fprintf('model::load_config: Loading configuration data %s\n', configfile);
                 fid=fopen(configfile);
                while 1
                    tline = fgetl(fid);
                    if ~ischar(tline), break, end
                    if ~isempty(tline) && ~strcmp(tline(1), '%') 
                        eval(['a.' tline]);
                    end
                end
                fclose(fid);
            end;
        end
        
        function s=adjust(a, s)
            % Weight a score with the adjustment weights
            if ~isempty(a.adjustment_weights)
                s=s*a.adjustment_weights(a.targetid);
            end
        end
        
        function a=record_history(a, q)
            % Record the score history
            s=q.score;
            e=q.ebar;
            if isempty(s), s=a.auc_history(length(a.auc_history)); end
            if isempty(e), e=a.ebar_history(length(a.ebar_history)); end
            a.auc_history=[a.auc_history s];
            a.ebar_history=[a.ebar_history e];
            a.cash_history=[a.cash_history a.initial_budget-a.current_budget];
        end
        
        function h=plot_learning_curve(a)
            % Last point extrapolation
            if a.auc_wo_queried_samples
                last_point=a.auc_history(length(a.auc_history));
            else
                last_point=1;
            end
            % Plot the learning curve in a log scale
            nm=upper( class(a));
            nm(nm=='_')=' ';
            h=figure('Name', nm); 
            cash_spent=[a.cash_history(1:length(a.cash_history)-1) a.initial_budget];
            plot(scale_axis(a, cash_spent), [a.auc_history last_point], ...
                 'r', 'LineWidth', 2);
            hold on
            plot(scale_axis(a, cash_spent), [a.auc_history last_point], ...
                 'ro', 'MarkerSize', 8, 'MarkerFaceColor', 'r');
            plot(scale_axis(a, [1 a.initial_budget]), [1 1]);
            plot(scale_axis(a, [a.initial_budget a.initial_budget]), [0.5 1]);
            plot(scale_axis(a, [1 a.initial_budget]), [0.5 0.5]);
            if ~a.auc_wo_queried_samples
                plot(scale_axis(a, [1 a.initial_budget]), [0.5 1], '-.');
            end
            if ~a.auc_wo_queried_samples
                tt='Red: ';
            else
                tt='';     
            end
            if ~isempty(a.current_score), tt=[tt nm ' (global score=' num2str(round(10000*a.current_score)/10000) ')']; end
            if ~a.auc_wo_queried_samples
                tt = [tt '. Blue: RAND'];
            end
            %if strcmp(a.axis_scale, 'linear')
            %	gs=1/3; % This is valid for the linear case
            %    tt=[tt ' (global score=' num2str(round(10000*gs)/10000) ')'];
            %end
            title(tt);
            switch a.axis_scale
                case 'log'
                    xlabel('Log_2(Number of labels queried)');
                case 'sqrt'
                    xlabel('Sqrt(Number of labels queried)');
                otherwise
                    xlabel('Number of labels queried');
            end
            ylabel('Area under the ROC curve');
            if ~a.auc_wo_queried_samples
                hold on
                % Rand algorithm expected results:
                cash_spent=1:a.initial_budget;
                plot(scale_axis(a, cash_spent), 1-0.5*(1-cash_spent/a.initial_budget).^2), 'b-';
            end
        end
        
        function a=compute_score(a)
            %a=compute_score(a)
            
            % Last point extrapolation
            if a.auc_wo_queried_samples
                last_point=a.auc_history(length(a.auc_history));
            else
                last_point=1;
            end
            x=scale_axis(a, [a.cash_history(1:length(a.cash_history)-1) a.initial_budget]);
            y=[a.auc_history last_point];
            % Compute the score
            s=0;
            % Integrate by the trapeze method
            for k=2:length(x)
                deltax=x(k)-x(k-1);
                mu=(y(k)+y(k-1))/2;
                s=s+deltax*mu;
            end
            % Normalize the score
            if a.auc_wo_queried_samples
                bias=scale_axis(a, a.initial_budget)*0.5;
                norma=bias;
            else
                
                bias=scale_axis(a, a.initial_budget)*0.75;
                norma=scale_axis(a, a.initial_budget)*0.25;
            end
            s=(s-bias)/norma;
            a.current_score=s;
        end
        
        function x=scale_axis(a, x)
            %x=scale_axis(a, x)
            % Scale the x axis with the axis_scale property
            switch a.axis_scale
                case 'log'
                    x=log2(x);
                case 'sqrt'
                    x=sqrt(x);
                % otherwise do nothing      
            end
        end
    end % protected
end
