%==========================================================================
% MODEL template for a data generative model          
%==========================================================================
% M=model(hyper) 
% hyper -- Hyperparameters (public properties). Use the Spider syntax.
%   train_num (Size of the training set) 
%   test_num (Size of the test set)
%   cost_per_sample
%   cost_per_target_observation 
%   cost_per_var_observation
%   cost_per_var_manipulation
% By default, all values are set to zero.
% If a configuration file model_config.txt exists, the hyperparameters are
% loaded from that file. Any hyperparameter specified as argument
% overwrites the default or configuration values.
%
% Examples:
%   default(model) % list the default values of the HP
%   M=model;
%   M=model('train_num=3000');
%   M=model({'train_num=3000', 'cost_per_sample=234'});
%   M.task_n_pricing or task_n_pricing(M)
%   B=initial_budget(M);
%   var_profile(M)  % shows the variable properties
%   [Q, M]=process_query(M, query);

% Isabelle Guyon -- isabelle@clopinet.com -- October 2009


classdef model
    % <<------------- hyperparameters ----------------->> 
    properties 
        train_num=[]; % load_config/eval_hyper: Size of the training set 
        test_num=[];  % load_config/eval_hyper: Size of the test set(s)

        cost_per_sample=0;              % load_config/eval_hyper 
        cost_per_target_observation=0;  % load_config/eval_hyper
        cost_per_var_observation=0;     % load_config/eval_hyper
        cost_per_var_manipulation=0;    % load_config/eval_hyper
    end
    % <<------------- model ----------------->> 
    properties (SetAccess = protected)
        time_dept=[];       % init: Time dependency (0/1)
        target=[];          % init: Indices of the target variables
        observable=[];      % init: Indices of the observable variables
        actionable=[];      % init: Indices of the actionable variables
        unobservable=[];    % init: Indices of the unobservable variables    
        train_cost=0;       % compute_budget: Price of the default labeled training set
        test_cost=0;        % compute_budget: Price of the default test set
        initial_budget=0;   % compute_budget: Initial budget
    end
    properties (GetAccess = protected)
        contact_name='Isabelle Guyon';
        contact_url='http://clopinet.com/isabelle/';
        resource_url=''; % URL of the web page of the model
        title='GLOP model';
        authors='Causality Workbench Team';
        keywords='bayesian.network, structural.equation.models, boolean.network';
        varprofile=[];
    end
    
    % <<------------- public methods ----------------->> 
    methods
        %%%%%%%%%%%%%%%%%%%
        %%% CONSTRUCTOR %%%
        %%%%%%%%%%%%%%%%%%%
        function a=model(hyper)
            fprintf('model::model: %s\n', upper(class(a)));
            % Load default hyperparameter values from a file
            a=load_config(a);
            % Spider script to evaluate hyperparam (overwrite defaults)
            eval_hyper;
            % Initialize the model or data (here just random values)
            % This must set the list of target, observable, actionable,
            % and unobservable variables; also set taim_dept.
            a=init(a);
            % Compute the price of the default datasets and initial budget.
            a=compute_budget(a);
            % Profile the variables
            a=var_profile(a);
        end 
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % IMPORTANT FUNCTIONS TO REDEFINE %
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % initial_budget (public)
        % task_n_pricing (public)    
        % process_query (public)
        % init (protected)
    end
    
    % <<------------- protected methods ----------------->> 
    methods (Access = protected)      
        function a=init(a)
            % Initialize the resourse name
            resource_name=upper(class(a));
            title=[resource_name 'GLOP model'];
            % Initialize datasets
            dataname=[class(a) '_dataset.mat'];
            if ~(exist([whereis(a) dataname], 'file')==2)
                fprintf('model::init: Initializing dataset %s\n', dataname);
                Xtr=randn(a.train_num, a.varnum);
                Ytr=sign(randn(a.train_num, a.targnum));
                Xte=randn(a.test_num, a.varnum);
                Yte=sign(randn(a.test_num, a.targnum));
                save([whereis(a) dataname], 'Xtr', 'Ytr', 'Xte', 'Yte');
            else
                fprintf('model::init: Using dataset %s\n', dataname);
            end
            a.time_dept=0; % No time dependency
            a.target=0;    % Indices of the target variables
            a.observable=0:11; % Indices of the observable variables
            a.actionable=1:11; % Indices of the actionable variables
            a.unobservable=[]; % Indices of the unobservable variables    
        end   
        function a=compute_budget(a)
            a.train_cost=a.train_num*(a.cost_per_sample+length(a.observable)*a.cost_per_var_observation+a.targnum*a.cost_per_target_observation); 
            % Price of the default test set           
            a.test_cost=a.test_num*(a.cost_per_sample+a.varnum*a.cost_per_var_observation); 
            % Initial budget
            a.initial_budget=a.train_cost+a.test_cost;      % This pays for the training and test sets 
        end
        function a=load_config(a)
            % Load hyperparamters from configuration file
            configfile=[class(a) '_config.txt'];
            if (exist(configfile, 'file')==2)
                fprintf('model::load_config: Loading configuration data %s\n', configfile);
                c= load(configfile, '-ASCII');
                a.train_num = c(1);
                a.test_num = c(2);
                a.cost_per_sample = c(3);
                a.cost_per_target_observation = c(4);
                a.cost_per_var_observation = c(5);
                a.cost_per_var_manipulation = c(6);
            end
        end
    end
    
    % <<------------- other simple public methods ----------------->>
    
    % Those need rarely changes
    methods
        function s=default(a, property)
            % s=default;
            % s=default(a, property)
            % Function that runs the constructor to determine the 
            % default values of hyperparameters (public properties).
            b=eval(class(a));
            names={'train_num','test_num','cost_per_sample','cost_per_target_observation','cost_per_var_observation','cost_per_var_manipulation'};
            if nargin<2
                for k=1:length(names)
                    s.(names{k})=b.(names{k});
                end
            else
            s=b.(property);
            end
        end
        
        function xml_display(a)
        %xml_display(a)
            exclude_list={'bnet'};
            fprintf('\t\t<model name="%s">\n', class(a));
            f=properties(a);
            for k=1:length(f)
                fn=f{k};
                val=a.(fn);
                if ( (isnumeric(val) || islogical(val)) && isempty(strmatch(fn, exclude_list))) 
                    if length(val)==1 && ~strcmp(fn, 'target')
                        fprintf('\t\t\t<%s value="%d" />\n', fn, val);
                    else 
                        fprintf('\t\t\t<num_%s value="%g" />\n', fn, length(val));
                    end
                end
            end
            fprintf('\t\t</model>\n');
        end
        
        function c=var_profile(a, num, as_xml, show)
            % a=var_profile(a)
            % c=var_profile(a, num)
            % Function that returns the profile of variables 
            % name, accessibility (observable, actionable, unobservable),
            % type (catagorical, binary, ordinal, continuous), min value,
            % max value. Note: unobservable variables are neither
            % observable nor actionable. Actionable variables are
            % observable.
            if nargin<2, num=[]; end
            if nargin<3, as_xml=0; end
            if nargin<4, show=0; end
            if any(a.target==0), add1=1; else add1=0; end
            if isempty(a.varprofile)
                header={'Index', 'Name', 'Access', 'Type', 'Min', 'Max'};
                name=var_name(a);
                access=var_access(a);
                type=var_type(a);
                min=var_min(a);
                max=var_max(a);
                N=length(name);
                c=cell(N,6);
                for k=1:N
                    c{k,1}=num2str(k-add1);
                    c{k,2}=name{k};
                    c{k,3}=access{k};
                    c{k,4}=type{k};
                    c{k,5}=num2str(min(k));
                    c{k,6}=num2str(max(k));
                end
                c=[header; c];
                a.varprofile=c;
            else
                c=a.varprofile;
            end
            if show,
                for i=1:size(c, 1)
                    for j=1:size(c, 2)
                        fprintf('%s\t', c{i, j});
                    end
                    fprintf('\n');
                end
            end
            if ~isempty(num),
                n=num+1+add1;
                if as_xml
                    c=sprintf('<variable index="%s" name="%s" access="%s" type="%s" min="%s" max="%s" />', ...
                        c{n, 1}, c{n, 2}, c{n, 3}, c{n, 4}, c{n, 5}, c{n, 6});
                else
                    c=c(n,:); 
                end
            else
                c=a;
            end
        end
        
        function save_profile(a, filename)
            %save_profile(a, filename)
            if nargin<2
                fp=2;
            else
                fp=fopen(filename,'w');
            end
            fprintf(fp, '<model name="%s">\n', class(a));
            fprintf(fp, '\t<contact_name value="%s" />\n',a.contact_name);
            fprintf(fp, '\t<contact_url value="%s" />\n', a.contact_url);
            fprintf(fp, '\t<resource_name value="%s" />\n', upper(class(a)));
            if ~isempty(a.resource_url), fprintf(fp, '\t<resource_url value="%s" />\n', a.resource_url); end
            fprintf(fp, '\t<title value="%s" />\n', a.title);
            fprintf(fp, '\t<authors value="%s" />\n',a.authors);
            fprintf(fp, '\t<keywords value="%s" />\n', a.keywords);
            warning off
            if isfield(struct(a), 'axis_scale')
                fprintf(fp, '\t<axis_scale value="%s" />\n', a.axis_scale);
            end
            warning on
            f=properties(a);
            exclude_list={'bnet', 'current_budget', 'targetid', 'adjustment_weights', 'queried_samples', 'cash_history', 'auc_history', 'ebar_history', 'current_score'};
            for k=1:length(f)
                fn=f{k};
                val=a.(fn);
                if ( (isnumeric(val) || islogical(val)) && (isempty(strmatch(fn, exclude_list)) || strcmp(fn, 'target')))
                    if length(val)==1 && ~strcmp(fn, 'target')
                        fprintf(fp, '\t<%s value="%d" />\n', fn, val);
                    else 
                        fprintf(fp, '\t<num_%s value="%d" />\n', fn, length(val));
                    end
                end
            end
            T=task_n_pricing(a);
            fprintf(fp, '\t<task>\n');
            fprintf(fp, '%s\n', T);
            fprintf(fp, '\t</task>\n\t<variables num="%d"><![CDATA[\n', varnum(a)+targnum(a));
            C=a.varprofile;
            for i=1:size(C, 1)
                for j=1:size(C, 2)-1
                    fprintf(fp, '%s\t', C{i, j});
                end
                fprintf(fp, '%s\n', C{i, size(C, 2)});
            end     
            fprintf(fp, ']]>\n');
            if any(a.target==0), base=0; else base=1; end
            for i=base:varnum(a)
                fprintf(fp, '%s\n', var_profile(a, i, 1));
            end
            fprintf(fp, '\t</variables>\n</model>');
            if fp>2, fclose(fp); end
        end     
    
        function p=whereis(a)
            %p=whereis(a)
            % Path to the object
            fn=which(class(a));
            dt=findstr('\', fn); 
            if ~isempty(dt), dt = dt(length(dt)); else dt=0; end
            dt2=findstr('/', fn); 
            if ~isempty(dt2), dt2 = dt2(length(dt2)); else dt2=0; end
            dt=max(dt, dt2);
            p=fn(1:dt);
        end      
        
        function n=varnum(a)
            %n=varnum(a)
            % number of variables, except the target
            n=length(setdiff(a.observable,a.target));
        end  
        function n=targnum(a)
            %n=targnum(a)
            % number of target(s)
            n=size(a.target, 2);
        end
        function v=var_name(a, num)
            %v=var_name(a, num)
            % Optionally supply a function that names the variables
            % Here we just name them with their order number
            if any(a.target==0), add1=1; else add1=0; end
            n=varnum(a); 
            v=cell(1,n); 
            for k=1:n, v{k}=['var' num2str(k)]; end
            v=['target', v];
            if nargin>1, v=v{num+add1}; end 
        end
        function v=var_access(a, num)
            %v=var_access(a, num)
            % Variable accessibility
            if any(a.target==0), add1=1; else add1=0; end
            n=varnum(a); 
            v=cell(1,n);
            [v{:}]=deal('unobservable');
            [v{a.observable+add1}]=deal('observable');
            [v{a.actionable+add1}]=deal('actionable');
            if nargin>1, v=v{num+add1}; end 
        end
        function v=var_type(a, num)
            %v=var_type(a, num)
            % Optionally supply a function providing variable type: binary,
            % categorical, ordinal, or continuous
            if any(a.target==0), add1=1; else add1=0; end
            n=varnum(a); 
            v=cell(1,n); 
            for k=1:n, v{k}='continuous'; end
            v=['binary' v];
            if nargin>1, v=v{num+add1}; end 
        end
        function v=var_min(a, num)
            %v=var_min(a, num)
            % Optionally supply a function that returns the minimum value
            % of variables
            if any(a.target==0), add1=1; else add1=0; end
            n=varnum(a); 
            v=-Inf*ones(1,n); 
            v=[-1 v];
            if nargin>1, v=v(num+add1); end 
       end
       function v=var_max(a, num)
            %v=var_max(a, num)
            % Optionally supply a function that returns the minimum value
            % of variables 
            if any(a.target==0), add1=1; else add1=0; end
            n=varnum(a); 
            v=Inf*ones(1,n); 
            v=[1 v];
            if nargin>1, v=v(num+add1); end 
       end       
    end

end




 

 
 





