function [q, a] = process_query(a, q)
%q = process_query(a, q)
%[q, a] = process_query(a, q)
% Inputs:
% a -- A model object.
% q -- A query object or a sample number (if observational data are
% requested only).
% Returns:
% q -- The query object, now containing the answer.
% a -- Optionally get back a with updated budget.

% Isabelle Guyon -- October 2009 -- isabelle@clopinet.com

dataname=[whereis(a) class(a) '_dataset.mat'];

if is_predict(q)
    % Compute the score and assign it to the query
    d=load(dataname);
    [s, e] = auc(postmanipval(q), d.Yte); % The AUC between predicted values and truth values (replace this with whatever score you want)
    % Note: here we drew target values at random: obviously the values should be
    % drawn from a reasonable model and take into account eventual
    % manipulations
    q=score(q, s, e); % Keep this: store the results (score and error bar) in the query
end

% Draw samples
% Important: update first the number of variables queried
q=quantity(q, a);
% and verify the range of the variables
q=verify_range(q, a);

% PREMANIPULATION DATA
% ====================
d=load(dataname); % See the constructor
if is_train(q) || is_survey(q)
    Xinit=[d.Ytr, d.Xtr];
elseif is_test(q)
    Xinit=[zeros(size(d.Yte)), d.Xte]; % Never return test labels
else
    Xinit=randn(samplenum(q), length(a.observable));
end

nidx0=premanipvar(q);
pidx=sample(q); % for survey data
if isempty(pidx), pidx=1:size(Xinit, 1); end
if ~isempty(nidx0)
    q=premanipval(q, Xinit(pidx, nidx0+1)); %nidx0: 0-based indices
end

% MANIPULATION DATA
% =================
if is_test(q)
    % Return the manipulations to be performed
    % The test set 0 is observational data
    % But we SET manipvar and manipval (this is our query to the participants)
    % For instance write a function
    % q=manipulate(a, q);
    % the ends with
    % q=manipvar(q, manip_var);
    % q=manipval(q, manip_val);
    % to set the values of q.manip_var and q.manip_val
end

% POSTMANIPULATION DATA
% ====================
% Create a manipulation matrix
nidx1=manipvar(q); 
nidx2=postmanipvar(q); 
do_manip = ~isempty(nidx1)&&~isempty(nidx2);
Xmanip=[];
if do_manip
    Xmanip=zeros(size(Xinit));
    Xmanip(:, nidx1+1)=manipval(q); % nidx1 is zero-based
end

Xresu=[];
N=length(setdiff(a.observable,a.target)); % number of variables (excluding the target)
if do_manip 
    q=varnum(q, N); 
    Xresu=zeros(size(Xinit));
    % Here to fill in values you must write some code
    %Xresu=draw_sample(a, Xinit, Xmanip);
    q=postmanipval(q, Xresu(:, nidx2+1), N); % nidx2 is zero-based
end

% COMPUTE THE COST OF THE QUERY
% =============================
% Re-evaluate the actual quantity (based on the results delivered)
q=quantity(q);
sample_cost=q.samplenum*a.cost_per_sample;
target_cost=q.targetnum*a.cost_per_target_observation;
observ_cost=q.obsernum*a.cost_per_var_observation;
manip_cost=0;
if ~is_test(q)
    manip_cost=q.manipnum*a.cost_per_var_manipulation;
end
total_cost=sample_cost+observ_cost+manip_cost+target_cost;
q=cost(q, total_cost, sample_cost, manip_cost, observ_cost, target_cost); % You must update the query cost

end

