%========================================================================================================  
% QUERY object
%========================================================================================================
% q = query;
% q = query(filename);
% q = query(filename, fn2);
%
% Stores query information and results.
% The argument "filename" is the root name of the query file.
% If fn2 is specified, make a copy to the second file.
% The query files may also be bundled in a zip archive, 
% in which case filename is the name of the archive.
%
% One expects a subset of the files:
% <filename>.type: Type of the query, types={'TRAIN', 'TEST', 'OBS',
% 'SURVEY', 'EXP', 'PREDICT'}; TEST and PREDICT can be followed by the test
% set number; OBS can be followed by the number of samples requested.
% <filename>.premanipvar, <filename>.manipvar, <filename>.postmanipvar: 
% One line with a list of variables. 
% <filename>.premanipval, <filename>.manipval, <filename>.postmanipval: 
% Multiline matrix, each line corresponding to one instance.
% Each column corresponds to a variable. 
% The query object stores this information in corresponding fields.
% Missing values are coded as NaN.
% Methods:
% load --       called by the constructor to load the this.
% save --       save the query to file
% struct --     show the structure
% xml_display -- show structure as XML
% In addition, many fields can be set/get with a method having the same
% name; see methods(query)

% Isabelle Guyon -- isabelle@clopinet.com -- May-Oct 2009

classdef query
    properties (SetAccess = private)
        % Identification
        participant_ID=[];
        model_name=[];
        experiment_name=[];
        date_submitted=[];
        type='';
        % Inputs
        sample=[];
        manipvar=[];
        manipval=[];
        postmanipvar=[];
        % Auxiliary fields and output fields
        postmanipval=[]; % The values for the queried "queryvar" variables
        premanipvar=[]; % The list of variables in the premanipval array
        premanipval=[]; % The values used to initialize the system before manipulating it
                   % This is used if paricular subjects are chosen on which
                   % to perform a given manipulation rather than drawing
                   % them at random according to the natural distribution.
                   % This is also used to return training or test data.
        predict=[];  % Same as postmanipval
        varnum=0; % Total number of variables (not including the target)
                 % ie the maximum variable index value
        samplenum=[]; % Number of samples asked for
        manipnum=[]; % Effective number of time a variable is manipulated in the query
        obsernum=[]; % Effective number of time a variable has been observed (value requested) in the query, including targets
        targetnum=[]; % Number of times a target variable is observed
        % Private
        types={'TRAIN', 'TEST', 'OBS', 'SURVEY', 'EXP', 'PREDICT', 'AL'};
        plot_handle=[];
    end
    properties 
        % All the costs are computed by the model
        samplecost=0; % samplenum * cost_per_sample
        manipcost=0; % manipnum * cost_per_manip
        obsercost=0; % obsernum * cost_per_observation
        targetcost=0;  % targetnum * cost_per_target (this is on top of the regular cost for observing a variable: the target may be more expensive
        totalcost=0; % Overall cost 
        % Prediction score and estimated error bar
        score=[];
        ebar=[];
        global_score=[];
        global_ebar=[];
        % Flag(s)
        is_overbudget=0; % Budget exceeded
        % Comments
        README='';
    end
    
    methods
        function this = query(filename, fn2) 
            if nargin>0
                sp=strfind(filename, ' ');
                if isempty(sp), sp=length(filename)+1; end
                key=filename(1:sp-1);
                if ~isempty(key) && ~isempty(strmatch(key, this.types))
                    this.type=filename;
                else
                    this=load(this, filename);
                end
            end
            this=sanitize(this);
            
            if nargin>1
                this.varnum=[]; 
                this.samplenum=[]; 
                this.manipnum=[]; 
                this.obsernum=[]; 
                this.samplecost=[];
                this.manipcost=[];
                this.obsercost=[];
                this.totalcost=[];  
                save(this, fn2, 1);
            end
        end
        
        function q=set_plot_handle(q, h)
            q.plot_handle=h;
        end
    end
    
    methods (Access=protected)
        function warning(a, message)
            fprintf('WARNING: %s\n', message);
        end
    end
end




