function a=sanitize(a)
%a=sanitize(a)
% Verify that a query is correct and remove forbidden parts.

% Isabelle Guyon -- isabelle@clopinet.com -- Oct 2009

sp=strfind(a.type, ' ');
if isempty(sp) || sp<=0
    sp=length(a.type);
else
    sp=sp-1;
end
mytype=a.type(1:sp);
myval=max(1, str2num(a.type(sp+1:length(a.type)))); % Values lower than 1 impossible

if isempty(mytype)
    if ~isempty(a.sample) && isempty(a.manipvar)
        if ~isempty(a.predict)
            mytype='AL'; % Special active learning query
            warning('query::sanitize: Missing query type, AL assumed');
        else
            mytype='SURVEY';
            warning('query::sanitize: Missing query type, SURVEY assumed');
        end
    elseif ~isempty(a.predict) || ~isempty(a.postmanipval) || ~isempty(a.premanipval)
        mytype='PREDICT';
        warning('query::sanitize: Missing query type, PREDICT assumed');
    else
        warning('query::sanitize: Missing query type, OBS assumed');
        mytype='OBS';
    end
    a.type=mytype;
elseif ~ismember(mytype, a.types)
    error('query::sanitize: Unknown query type %s\n', mytype);
end

compulsory_list=[];
switch mytype
    case {'TRAIN', 'TEST', 'OBS'}
        ignore_list={'sample', 'manipvar', 'postmanipvar', 'premanipval', 'manipval', 'postmanipval', 'predict'};
    case 'EXP'       
        ignore_list={'premanipval', 'postmanipval'};
        compulsory_list={'manipvar', 'postmanipvar', 'manipval'};
    case 'SURVEY'
        ignore_list={'manipvar', 'postmanipvar', 'premanipval', 'manipval', 'postmanipval', 'predict'};
        compulsory_list={'sample'};
        if isempty(a.premanipvar)
            warning('query::sanitize: SURVEY query, by default premanipvar set to 0');
            a.premanipvar=0;
        end
    case 'PREDICT'
        ignore_list={'sample', 'premanipvar', 'manipvar', 'premanipval', 'manipval'};
        if ~isempty(a.predict) 
            if ~isempty(a.postmanipval) 
                warning('query::sanitize: PREDICT query in which both predict and postmanipval are provided; only predict is taken into account\n');
            end
            a.postmanipval=a.predict;
            a.predict=[];   
        end
        if isempty(a.predict) && isempty(a.postmanipval) && ~isempty(a.premanipval)
            warning('query::sanitize: PREDICT query in which premanipval is provided but not predict or postmanipval; we assume that premanipval contains the predictions\n');
            a.postmanipvar=a.premanipvar;
            a.postmanipval=a.premanipval;
            a.predict=[];
        end
        % By default, postmanipvar is the target variable
        if isempty(a.postmanipvar)
            a.postmanipvar=0;
        end
        compulsory_list={'postmanipval'};
    case 'AL'
        ignore_list={'manipvar', 'premanipval', 'manipval'};
        if ~isempty(a.predict) 
            if ~isempty(a.postmanipval) 
                warning('query::sanitize: AL query in which both predict and postmanipval are provided; only predict is taken into account\n');
            end
            a.postmanipval=a.predict;
            a.predict=[];   
        end
        % By default, postmanipvar is the target variable, for predict
        % queries
        if isempty(a.postmanipvar)
            warning('query::sanitize: AL query, by default postmanipvar set to 0');
            a.postmanipvar=0;
        end
        % By default, postmanipvar is the target variable, for survey
        % queries
        if isempty(a.premanipvar)
            warning('query::sanitize: AL query, by default premanipvar set to 0');
            a.premanipvar=0;
        end
        compulsory_list={'sample', 'postmanipval'};
    otherwise
        error('query::sanitize: Unknown query type %s\n', mytype);
end

% Delete forbidden fields
for k=1:length(ignore_list)
    if ~isempty(a.(ignore_list{k}))
        warning(['query::sanitize: ' a.type ' query. Extra ' ignore_list{k} ' file provided; ignored']);
        a.(ignore_list{k})=[];
    end
end

% Mandatory fields
for k=1:length(compulsory_list)
    if isempty(a.(compulsory_list{k}))
        warning(['query::sanitize: ' a.type ' query. Missing ' compulsory_list{k} ' file!']);
        a.README = [a.README ' WARNING: Missing ' compulsory_list{k} ' file!'];
    end
end

% Verify that the number of samples is specified correctly
if is_observation(a) 
    if isempty(a.premanipvar)
        warning('query::sanitize: Missing premanipulation variables, will assume 1:N');       
    end
end
if size(a.sample,2)>1
    error('query::sanitize: sample file should contain a column vector');
end
if ~isempty(a.manipval) && ~isempty(a.sample)
    if size(a.sample,1)~=size(a.manipval,1)
        error('query::sanitize: Inconsistent samples sizes in sample and manipval files');
    end
end
    
% Verify the consistency of the number of variables in var and val files
if length(a.manipvar) ~= size(a.manipval, 2)
    error('query::sanitize: Wrong manip var dimensions');
end
if ~isempty(a.premanipval) % Double check
    error('query::sanitize: premanipval files cannot be specified, use sample files to specify pre-manipulation variables');
end

% Verify the sanity of the indices
varlist={'manipvar', 'premanipvar', 'postmanipvar'};
for k=1:length(varlist)
    if size(a.(varlist{k}), 1)>1
        error('query::sanitize: File %s should be a line vector\n', varlist{k});
    end
    if any(a.(varlist{k})<0)
        error('query::sanitize: Negative index in %s\n', varlist{k});
    end
    if length(unique(a.(varlist{k})))~=length(a.(varlist{k}))
        error('query::sanitize: Repeated index in premanipvar');
    end
end
% Cannot manipulate the target
idx=a.manipvar;
tidx=find(idx==0);
n=length(idx);
if ~isempty(tidx)
    warning('query::sanitize: Cannot manipulate the target; index 0 removed');
    idx=idx([1:tidx-1,tidx+1:n]);
end
a.manipvar=idx;

% Preliminary calculation of the number of samples
a=quantity(a);

function warning(message)
fprintf('%s\n', message);

